module decoder(pc, ir, state);
   input [15:0]  pc;
   input [15:0]  ir;		// instruction
   output [65:0] state;		// packed to pipeline state register
   wire [3:0] 	 opcode;	// instruction [15:12]
   wire [2:0] 	 DRa;		// DR register addr
   wire [2:0] 	 SR1a;		// SR1 register addr
   wire [2:0] 	 SR2a;		// SR2 register addr
   reg [2:0] 	 alu_ops;	// ALU operation: add, and, not
   wire 	 iswb1, iswb2;	// whether to write back to register
   wire [4:0] 	 sjnzp;		// PC => R7, unconditional jump, branch if nzp
   reg [15:0] 	 imm;		// sign extended immediate
   wire 	 PC_or_SR1;	// 0 for PC, 1 for SR1
   reg 		 imm_or_SR2;	// 0 for imm, 1 for SR2
   wire 	 memw;		// memory write
   wire 	 apsr_is_alu;	// write apsr from alu
   wire 	 apsr_is_mem;	// write apsr from memory (at most one high for these two wires)
   wire 	 load_mem;	// reading from memory to register
   wire 	 arith_alu;	// ALU is doing arithmetic operation
   wire 	 indirect_mem;	// LDI or STI
   wire [15:0] 	 added_pc; 	// PC should increment 1 when saved
   wire 	 trap;		// TRAP
   wire		 exc;		// exception
   wire		 rti;		// Return from interrupt

   assign added_pc = pc + 1;
   state_pack decoder_pack(.state(state),
			   .pc(added_pc),
			   .op(opcode),
			   .dra(DRa),
			   .sr1a(SR1a),
			   .sr2a(SR2a),
			   .iswb_alu(iswb1),
			   .iswb_mem(iswb2),
			   .imm(imm),
			   .alu_ops(alu_ops),
			   .c_sr1(PC_or_SR1),
			   .c_sr2(imm_or_SR2),
			   .sjnzp(sjnzp),
			   .memw(memw),
			   .apsr_is_alu(apsr_is_alu),
			   .apsr_is_mem(apsr_is_mem),
			   .load_mem(load_mem),
			   .arith_alu(arith_alu),
			   .indirect_mem(indirect_mem),
			   .trap(trap),
			   .exc(exc),
			   .rti(rti)
			   );

   assign opcode = ir[15:12];
   assign DRa = ir[11:9];
   assign SR1a = ir[8:6];
   // if inst is in STORE series, then SR (write_data) = SR2a
   assign SR2a = memw ? ir[11:9] : ir[2:0];
   // alu_ops is in case(opcode)

   // iswb1 = ADD | AND | LEA | NOT
   assign iswb1 = (opcode == 4'b0001 || opcode == 4'b0101 ||
		   opcode == 4'b1110 || opcode == 4'b1001);

   // iswb2 = LD | LDI | LDR
   assign iswb2 = (opcode == 4'b0010 || opcode == 4'b1010 || opcode == 4'b0110);

   // brnzp is available only in BR
   assign sjnzp[2:0] = (opcode == 4'b0000) ? ir[11:9] : 3'b000;

   // unconditional jump: JMP | JSR | JSRR | TRAP | RTI
   assign sjnzp[3] = (opcode == 4'b1100 || opcode == 4'b0100 || opcode == 4'b1111 || opcode == 4'b1000);

   // PC => R7: JSR | JSRR | TRAP
   assign sjnzp[4] = (opcode == 4'b0100 || opcode == 4'b1111);

   // exception
   assign exc = (opcode == 4'b1101);
   // RTI
   assign rti = (opcode == 4'b1000);
   
   // imm is in case(opcode)

   // PC_or_SR1 = ADD | AND | JMP | JSRR | LDR | NOT | STR | RTI
   assign PC_or_SR1 = (opcode == 4'b0001 || opcode == 4'b0101 ||
		       opcode == 4'b1100 || opcode == 4'b0100 && ~ir[11] ||
		       opcode == 4'b0110 || opcode == 4'b1001 ||
		       opcode == 4'b0111 || opcode == 4'b1000);
   // imm_or_SR2 is in case(opcode)

   // memw = ST | STI | STR
   assign memw = (opcode == 4'b0011 || opcode == 4'b1011 || opcode == 4'b0111);
   // apsr_is_alu = ADD | AND | NOT | LEA
   assign apsr_is_alu = (opcode == 4'b0001 || opcode == 4'b0101 || opcode == 4'b1001 || opcode == 4'b1110);
   // apsr_is_mem = LD | LDI | LDR
   assign apsr_is_mem = (opcode == 4'b0010 || opcode == 4'b1010 || opcode == 4'b0110);

   // load_mem = LD | LDI | LDR
   assign load_mem = (opcode == 4'b0010 || opcode == 4'b1010 || opcode == 4'b0110);
   // arith_alu = ADD | AND | NOT
   assign arith_alu = (opcode == 4'b0001 || opcode == 4'b0101 || opcode == 4'b1001);
   // indirect_mem = STI | LDI
   assign indirect_mem = (opcode == 4'b1011 || opcode == 4'b1010);
   // trap = TRAP | EXCEPTION
   assign trap = (opcode == 4'b1111 || opcode == 4'b1101);

   always@ (ir)
     case (ir[15:12]) // opcode
       4'b0001: // ADD
	 begin
	    alu_ops = 3'b100;
	    if (ir[5]) // DR = SR1 + SEXT(imm5)
	      begin
		 if (~ir[4])
		   imm[15:5] = 11'b00000000000;
		 else
		   imm[15:5] = 11'b11111111111;
		 imm[4:0] = ir[4:0];
		 imm_or_SR2 = 0;
	      end
	    else		// DR = SR1 + SR2
	      begin
		 imm_or_SR2 = 1;
		 imm = 0;
	      end
	 end
       4'b0101: // AND
	 begin
	    alu_ops = 3'b010;
	    if (ir[5]) // DR = SR1 & SEXT(imm5)
	      begin
		 if (~ir[4])
		   imm[15:5] = 11'b00000000000;
		 else
		   imm[15:5] = 11'b11111111111;
		 imm[4:0] = ir[4:0];
		 imm_or_SR2 = 0;
	      end
	    else		// DR = SR1 & SR2
	      begin
		 imm_or_SR2 = 1;
		 imm = 0;
	      end
	 end
       4'b0000: // BR
	 begin	// PC = PC + SEXT(PCoffset9)
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b0000000;
	    else
	      imm[15:9] = 7'b1111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b1100: // JMP
	 begin
	    alu_ops = 3'b000;
	    imm = 0;
	    imm_or_SR2 = 0; // no need
	 end
       4'b0100: // JSR/JSRR
	 begin
	    if (ir[11])	// JSR: PC = PC + SEXT(PCoffset11)
	      begin
		 alu_ops = 3'b100;
		 if (~ir[10])
		   imm[15:11] = 5'b00000;
		 else
		   imm[15:11] = 5'b11111;
		 imm[10:0] = ir[10:0];
		 imm_or_SR2 = 0;
	      end
	    else			// JSRR: PC = BaseR (do not need addition)
	      begin
		 alu_ops = 3'b000;
		 imm_or_SR2 = 0;
		 imm = 0;
	      end
	 end
       4'b0010: // LD
	 begin	// DR = M(PC + SEXT(PCoffset9))
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b0000000;
	    else
	      imm[15:9] = 7'b1111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b1010: // LDI
	 begin	// DR = M(M(PC + SEXT(PCoffset9)))
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b0000000;
	    else
	      imm[15:9] = 7'b1111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b0110: // LDR
	 begin	// DR = M(BaseR + SEXT(PCoffset6))
	    alu_ops = 3'b100;
	    if (~ir[5])
	      imm[15:6] = 10'b0000000000;
	    else
	      imm[15:6] = 10'b1111111111;
	    imm[5:0] = ir[5:0];
	    imm_or_SR2 = 0;
	 end
       4'b1110: // LEA
	 begin	// DR = PC + P9
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b0000000;
	    else
	      imm[15:9] = 7'b1111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b1001: // NOT
	 begin	// DR = ~SR
	    alu_ops = 3'b001;
	    imm = 0;
	    imm_or_SR2 = 0;
	 end
       /*		4'b1100: // RET
	begin	// PC = R7
	alu_ops = 3'b000;
	imm = 0;
	imm_or_SR2 = 0;	// no need
			end
	*/
       4'b1000: // RTI
	 begin	// PC = M(R6)
	    alu_ops = 3'b000;
	    imm = 0;
	    imm_or_SR2 = 0;
	 end
       4'b0011: // ST
	 begin	// M(PC + P9) = SR
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b000000;
	    else
	      imm[15:9] = 7'b111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b1011: // STI
	 begin	// M(M(PC + P9)) = SR
	    alu_ops = 3'b100;
	    if (~ir[8])
	      imm[15:9] = 7'b0000000;
	    else
	      imm[15:9] = 7'b1111111;
	    imm[8:0] = ir[8:0];
	    imm_or_SR2 = 0;
	 end
       4'b0111: // STR
	 begin	// M(BaseR + SEXT(offset6)) = SR
	    alu_ops = 3'b100;
	    if (~ir[5])
	      imm[15:6] = 10'b0000000000;
	    else
	      imm[15:6] = 10'b1111111111;
	    imm[5:0] = ir[5:0];
	    imm_or_SR2 = 0;
	 end
       4'b1111: // TRAP
	 begin	// R7 = PC, PC = M[trapvect8]
	    alu_ops = 3'b000;
	    imm = {8'b00000000, ir[7:0]};
	    imm_or_SR2 = 0;
	 end
       default: // exception (4'b1101)
	 begin
	    alu_ops = 3'b000;
	    imm = 0;
	    imm_or_SR2 = 0;
	 end
     endcase

endmodule // decoder
